from pydantic import BaseModel,Field
from typing import ForwardRef,Union
from typing import List
from zt_backend.models.state.user_state import UserContext

Row = ForwardRef('Row')
Column = ForwardRef('Column')

class Column(BaseModel):
    """A Column must be a subcomponent of a Row. It can contain both individual components and rows"""
    components: List[Union[str,"Row"]] = Field([], description="List of component IDs and rows that belong to this column, rendered in order")
    width: Union[int,bool] = Field(False, description="Width of the column. It can be a number 1-12 and by default is automatically calculated")

    def __init__(__pydantic_self__, *args, **data):
        if args:
            data['components'] = args[0]
            if len(args) > 1:
                data['width'] = args[1]
        super().__init__(**data)
    
class Row(BaseModel):
    """Rows can contain both individual components and columns. They are the top level components of a layout and can be subcomponents of columns"""
    components: List[Union[str,Column]] = Field([], description="List of component IDs and columns that belong to this row, rendered in order")

    def __init__(self, *args, **data):
        if args:
            data['components'] = args[0]
        super().__init__(**data)

class Layout(BaseModel):
    """Layout is an object that contains the list of rows to be rendered"""
    rows: List[Row] = Field([], description="List of rows in this layout")

    def __init__(self, *args, **data):
        if args:
            data['rows'] = args[0]
        super().__init__(**data)
        execution_state = UserContext.get_state()
        if execution_state and execution_state.context_globals['exec_mode']:
            execution_state.current_cell_layout.append(self)


from pydantic import Field, validator,field_validator
from zt_backend.models.components.zt_component import ZTComponent
from zt_backend.models.components.validations import validate_color
from typing import List, Union, Optional
from zt_backend.models.state.user_state import UserContext

class SelectBox(ZTComponent):
    """SelectBox component allows users to select from a list of options. Can be a single or multiple selection"""
    
    component: str = Field("v-select", description="Vue component name")
    items: List[Union[str, int]] = Field(..., description="Options for the select box. Can be a list of strings or integers")
    value: Union[List[Union[str, int,None]],None,str,int] = Field(None, description="Selected options for the select box")
    label: Optional[str] = Field(None, description="Label for the select box")
    multiple: Optional[bool] = Field(None, description="Determines if multiple selections are allowed")
    dense: Optional[bool] = Field(None, description="Determines if the select box is dense")
    outlined: Optional[bool] = Field(None, description="Determines if the select box has an outlined style")
    clearable: Optional[bool] = Field(None, description="Determines if the select box has a clearable option")
    disabled: Optional[bool] = Field(None, description="Determines if the select box is disabled")
    readonly: Optional[bool] = Field(None, description="Determines if the select box is read-only")
    color: Optional[str] = Field(None, pre=True, description="Color of the range slider. Can be custom or standard Material color")
    triggerEvent: str = Field('update:modelValue',description="Trigger event for when to trigger a run")

    @field_validator('color')
    def validate_color(cls, color):
        return validate_color(color)

    @validator('value', always=True) #TODO: debug and replace with field validator
    def get_value_from_global_state(cls, value, values):
        id = values['id'] # Get the id if it exists in the field values
        execution_state = UserContext.get_state()
        try:
            if execution_state and id and id in execution_state.component_values:  # Check if id exists in global_state
                return execution_state.component_values[id]  # Return the value associated with id in global_state
        except Exception as e:
            e
        return value  # If id doesn't exist in global_state, return the original value
    
    @classmethod
    def from_dict(cls, items_map: dict, id: str, value: Union[str, int, None] = None, label: Optional[str] = None, multiple: Optional[bool] = None,
                   dense: Optional[bool] = None, outlined: Optional[bool] = None, clearable: Optional[bool] = None, disabled: Optional[bool] = None, 
                   readonly: Optional[bool] = None, color: Optional[str] = None, triggerEvent: str = 'update:modelValue'):
        """Creates a SelectBox instance from a dictionary."""
        items = items_map.keys()
        return cls(items=items, id=id, value=value, label=label, multiple=multiple, dense=dense, outlined=outlined, clearable=clearable, 
                   disabled=disabled, readonly=readonly, color=color, triggerEvent=triggerEvent)


from copilot.copilot import text_document_did_change
import traceback
import jedi
import logging

logger = logging.getLogger("__name__")

async def get_code_completions(cell_id:str, code: str, line: int, column: int) -> list:

    await text_document_did_change({
        "textDocument": {
            "uri": "file:///notebook.ztnb",
            "version": 1
        },
        "contentChanges": [{
            "text": code
        }]
    })

    code_list = code.split("\n")
    try:
        last_line = code_list[max(line-1,0)]
        if last_line:
            last_char = last_line[max(column-1,0)]
        else:
            last_char = ''
        if last_char in ['.', '(', '[', '{']:
            return {"cell_id": cell_id, "completions": []}

    except Exception as e:
        logger.info("Error formatting completions for cell_id %s: %s", cell_id, traceback.format_exc())

    try:
        script = jedi.Script(code)
        completions = script.complete(line, column)
        completions = [{"label": completion.name, "type": completion.type, "inlineFlag": False} for completion in completions]
        return {"cell_id": cell_id, "completions": completions}
    except Exception:
        logger.debug("Error getting completions for cell_id %s: %s", cell_id, traceback.format_exc())
        return {"cell_id": cell_id, "completions": []}


from fastapi import WebSocket

class ConnectionManager:
    def __init__(self):
        self.active_connections: list[WebSocket] = []

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket):
        self.active_connections.remove(websocket)

    async def send_text(self, message: str, websocket: WebSocket):
        await websocket.send_text(message)

    async def broadcast(self, message: str):
        for connection in self.active_connections:
            await connection.send_text(message)


from pydantic import Field
from zt_backend.models.components.zt_component import ZTComponent
from plotly.graph_objs import Figure
from plotly.io import to_json

def figure_to_json(figure: Figure) -> str:
    """Serializes a Plotly figure to a JSON string."""
    return to_json(figure, validate=False)

class PlotlyComponent(ZTComponent):
    """Component to display a Plotly figure"""
    component: str = Field("plotly-plot", description="Vue component name for Plotly")
    figure_json: str = Field(..., description="Serialized Plotly figure as JSON string")
    id: str = Field(..., description="ID for the component")

    @classmethod
    def from_figure(cls, figure: Figure, id: str):
        """Creates a PlotlyComponent instance from a Plotly figure."""
        figure.update_layout(template='plotly_dark')
        if not isinstance(figure, Figure):
            raise ValueError("Input must be a Plotly Figure")
        figure_json = figure_to_json(figure)  # Serialize figure to JSON
        return cls(figure_json=figure_json, id=id)

    def to_json(self):
        """Converts the component to a JSON-serializable dictionary."""
        return {
            "component": self.component,
            "figureJson": self.figure_json,
            "id": self.id
        }


from zt_backend.models.state.user_state import UserContext

class State(dict):
    @staticmethod
    def get_or_create_state(*args, **kwargs):
        current_state = UserContext.get_state()
        if current_state is not None and current_state.var_state is not None:
            return current_state.var_state
        else:
            new_state = State(*args, **kwargs)
            if current_state is not None:
                current_state.var_state = new_state
            return new_state

    def __setattr__(self, key, value):
        if key in self:
            super().__setattr__(key, value)
        else:
            raise AttributeError("Cannot add new attributes. Please use dictionary key-value pairs.")

    def __delattr__(self, key):
        if key in self:
            super().__delattr__(key)
        else:
            raise AttributeError("Cannot delete attributes. Please use dictionary operations.")
        
def state():
    return State.get_or_create_state()

global_state = State()


from typing import Union
from pydantic import Field, validator
from zt_backend.models.components.zt_component import ZTComponent
from zt_backend.models.state.user_state import UserContext

class iFrame(ZTComponent):
    """This is a representation of a standard iframe component"""
    component: str = Field("iframe", description="Vue component name")
    src: str = Field('', description="Source URL of the iframe")
    width: Union[int,str] = Field('100%', description="Width of the iframe")
    height: Union[int,str] = Field(300, description="Height of the iframe")
    frameborder: int = Field(0, description="Frame border of the iframe")
    scrolling: str = Field('auto', description="Scrolling of the iframe")
    allowtransparency: bool = Field(False, description="Allow transparency of the iframe")


from typing import OrderedDict
from zt_backend.models.state.user_state import UserState
from zt_backend.models.state.notebook_state import NotebookState
from zt_backend.models.api import request, response
from zt_backend.models import notebook
from zt_backend.utils.debounce import debounce
from zt_backend.config import settings
import logging
import duckdb
import uuid
import os
import traceback
import json
import copy
import rtoml

logger = logging.getLogger("__name__")
notebook_state = NotebookState()

def get_notebook(id=''):
    if id!='':
        try:
            logger.debug("Getting notebook from db with id %s", id)
            #get notebook from the database
            notebook_state.zt_notebook = get_notebook_db(id)
            notebook_state_init()
            return
        except Exception as e:
            logger.debug("Error when getting notebook %s from db: %s", id, traceback.format_exc())

    try:
        logger.debug("Notebook id is empty")            
        # If it doesn't exist in the database, load it from the TOML file
        with open('notebook.ztnb', "r") as project_file:
            toml_data = rtoml.loads(project_file.read().replace('\\','\\\\'))

        try:
            #get notebook from the database
            notebook_state.zt_notebook = get_notebook_db(toml_data['notebookId'])
            logger.debug("Notebook retrieved from db with id %s", toml_data['notebookId'])
            notebook_state_init()
            return
        except Exception as e:
            logger.debug("Error loading notebook with id %s from db: %s", toml_data['notebookId'], traceback.format_exc())
            pass
        # Convert TOML data to a Notebook object
        notebook_data = {
            'notebookId' : toml_data['notebookId'],
            'notebookName' : toml_data.get('notebookName', 'Zero True'),
            'userId' : '',
            'cells': {
                cell_id: notebook.CodeCell(id=cell_id, **cell_data, output="")
                for cell_id, cell_data in toml_data['cells'].items()
            }
        }
        notebook_state.zt_notebook = notebook.Notebook(**notebook_data)
        new_notebook = notebook_state.zt_notebook.model_dump_json()
        conn = duckdb.connect(notebook_state.notebook_db_path)
        create_table_query = f"CREATE TABLE IF NOT EXISTS '{notebook_state.zt_notebook.notebookId}' (id STRING PRIMARY KEY, notebook STRING)"
        conn.execute(create_table_query)
        insert_query = f"INSERT OR REPLACE INTO '{notebook_state.zt_notebook.notebookId}' (id, notebook) VALUES (?, ?)"
        conn.execute(insert_query, [notebook_state.zt_notebook.notebookId, new_notebook])
        conn.close()
        logger.debug("Notebook with id %s loaded from toml and new db entry created", toml_data['notebookId'])
        notebook_state_init()
    except Exception as e:
        logger.error("Error when loading notebook, return empty notebook: %s", traceback.format_exc())

def notebook_state_init():
    cells = []
    components={}
    for cell_key, cell in notebook_state.zt_notebook.cells.items():
        cell_request=request.CodeRequest(
            id=cell.id, 
            code=cell.code,
            variable_name=cell.variable_name,
            nonReactive=cell.nonReactive,
            showTable=cell.showTable,
            cellType=cell.cellType
        )
        for comp in cell.components:
            if hasattr(comp, 'value'):
                components[comp.id] = comp.value
        cells.append(cell_request)
    notebook_state.base_cells = cells
    notebook_state.base_components = components
    if settings.run_mode=='app':
        for cell_key, cell in notebook_state.zt_notebook.cells.items():
            if cell.hideCell or cell.hideCode:
                notebook_state.zt_notebook.cells[cell_key].code = ''

def get_notebook_request():
    return notebook_state.zt_notebook

def get_request_base(origin_id, components=None):
    if components is None:
        return request.Request(
                    originId=origin_id,
                    cells=copy.deepcopy(notebook_state.base_cells),
                    components=copy.deepcopy(notebook_state.base_components)
                )
    else:
        return request.Request(
                    originId=origin_id,
                    cells=copy.deepcopy(notebook_state.base_cells),
                    components=components
                )

def get_notebook_db(id=''):
    conn = duckdb.connect(notebook_state.notebook_db_path)
    if id!="":
        get_notebook_query = f"SELECT notebook FROM '{id}' WHERE id = '{id}'"
        notebook_data = conn.execute(get_notebook_query).fetchall()
        conn.close()
    return notebook.Notebook(**json.loads(notebook_data[0][0]))

def globalStateUpdate(newCell: notebook.CodeCell=None, 
                      position_key:str=None, 
                      deletedCell: str=None, 
                      saveCell: request.SaveRequest=None, 
                      hideCell: request.HideCellRequest=None,
                      hideCode: request.HideCodeRequest=None,
                      expandCode: request.ExpandCodeRequest=None,
                      renameCell: request.NameCellRequest=None,
                      cellReactivity: request.CellReactivityRequest=None,
                      showTable: request.ShowTableRequest=None,
                      run_request: request.Request=None, 
                      run_response: response.Response=None, 
                      new_notebook_name: str="",
                      ):
    logger.debug("Updating state for notebook %s", notebook_state.zt_notebook.notebookId)
    try:        
        old_state = notebook_state.zt_notebook.model_dump()
        if newCell is not None:
            if position_key:
                new_cell_dict = OrderedDict()
                for k, v in notebook_state.zt_notebook.cells.items():
                    new_cell_dict[k] = v
                    if k==position_key:
                        new_cell_dict[newCell.id] = newCell
                notebook_state.zt_notebook.cells = new_cell_dict
            else:
                new_cell_dict = OrderedDict({newCell.id: newCell})
                new_cell_dict.update(notebook_state.zt_notebook.cells)
                notebook_state.zt_notebook.cells = new_cell_dict
        if deletedCell is not None:
            del notebook_state.zt_notebook.cells[deletedCell]
        if saveCell is not None:
            notebook_state.zt_notebook.cells[saveCell.id].code=saveCell.text
        if hideCell is not None:
            notebook_state.zt_notebook.cells[hideCell.cellId].hideCell=hideCell.hideCell
        if hideCode is not None:
            notebook_state.zt_notebook.cells[hideCode.cellId].hideCode=hideCode.hideCode
        if expandCode is not None:
            notebook_state.zt_notebook.cells[expandCode.cellId].expandCode=expandCode.expandCode
        if renameCell is not None:
            notebook_state.zt_notebook.cells[renameCell.cellId].cellName=renameCell.cellName
        if cellReactivity is not None:
            notebook_state.zt_notebook.cells[cellReactivity.cellId].nonReactive=cellReactivity.nonReactive
        if showTable is not None:
            notebook_state.zt_notebook.cells[showTable.cellId].showTable=showTable.showTable
        if run_request is not None:
            for requestCell in run_request.cells:
                #zt_notebook.cells[requestCell.id].code = requestCell.code
                notebook_state.zt_notebook.cells[requestCell.id].variable_name = requestCell.variable_name
        if run_response is not None:
            for responseCell in run_response.cells:
                notebook_state.zt_notebook.cells[responseCell.id].components = responseCell.components
                notebook_state.zt_notebook.cells[responseCell.id].output = responseCell.output
                notebook_state.zt_notebook.cells[responseCell.id].layout = responseCell.layout
        if new_notebook_name:
            notebook_state.zt_notebook.notebookName = new_notebook_name
        save_notebook()
    except Exception as e:
        logger.error("Error while updating state for notebook %s: %s", notebook_state.zt_notebook.notebookId, traceback.format_exc())

@debounce(5)
def save_notebook():
    new_notebook = notebook_state.zt_notebook.model_dump_json()
    conn = duckdb.connect(notebook_state.notebook_db_path)
    insert_query = f"INSERT OR REPLACE INTO '{notebook_state.zt_notebook.notebookId}' (id, notebook) VALUES (?, ?)"
    conn.execute(insert_query, [notebook_state.zt_notebook.notebookId, new_notebook])
    conn.close()
    write_notebook()

def write_notebook():
    tmp_uuid_file = f'notebook_{uuid.uuid4()}.ztnb'
    logger.debug("Saving toml for notebook %s", notebook_state.zt_notebook.notebookId)
    try:
        with open(tmp_uuid_file, "w") as project_file:
            # Write notebookId
            project_file.write(f'notebookId = "{notebook_state.zt_notebook.notebookId}"\nnotebookName = "{notebook_state.zt_notebook.notebookName}"\n\n')

            for cell_id, cell in notebook_state.zt_notebook.cells.items():
                project_file.write(f'[cells.{cell_id}]\n')
                project_file.write(f'cellName = "{cell.cellName}"\n')
                project_file.write(f'cellType = "{cell.cellType}"\n')
                project_file.write(f'hideCell = "{cell.hideCell}"\n')
                project_file.write(f'hideCode = "{cell.hideCode}"\n')
                project_file.write(f'expandCode = "{cell.expandCode}"\n')
                project_file.write(f'showTable = "{cell.showTable}"\n')
                project_file.write(f'nonReactive = "{cell.nonReactive}"\n')
                
                if cell.cellType=='sql':
                    if cell.variable_name:
                        project_file.write(f'variable_name = "{cell.variable_name}"\n')
                
                # Format code as a multi-line string
                escaped_code = cell.code.encode().decode('unicode_escape').replace('"""',"'''")
                project_file.write(f'code = """\n{escaped_code}"""\n\n')
        os.replace(tmp_uuid_file, 'notebook.ztnb')
    except Exception as e:
        logger.error("Error saving notebook %s toml file: %s", notebook_state.zt_notebook.notebookId, traceback.format_exc())
        
    finally:
        try:
            os.remove(tmp_uuid_file)
        except Exception as e:
            logger.debug("Error while deleting temporary toml file for notebook %s: %s", notebook_state.zt_notebook.notebookId, traceback.format_exc())
            pass  # Handle error silently
    logger.debug("Toml saved for notebook %s", notebook_state.zt_notebook.notebookId)

async def save_worker(save_queue):
    while True:
        message = await save_queue.get()
        globalStateUpdate(**message)

async def websocket_message_sender(execution_state: UserState):
    while True:
        message = await execution_state.message_queue.get()
        await execution_state.websocket.send_json(message)


from threading import Timer
import asyncio

def debounce(wait):
    """ Decorator that will postpone a functions
        execution until after wait seconds
        have elapsed since the last time it was invoked. """
    def decorator(fn):
        def debounced(*args, **kwargs):
            def call_it():
                fn(*args, **kwargs)
            try:
                debounced.t.cancel()
            except(AttributeError):
                pass
            debounced.t = Timer(wait, call_it)
            debounced.t.start()
        return debounced
    return decorator


def async_debounce(wait):
    """Decorator that postpones an async function's execution until after `wait` seconds have elapsed since the last time it was invoked."""
    def decorator(fn):
        current_task = None

        async def debounced(*args, **kwargs):
            nonlocal current_task

            # Cancel the previous task if it exists and is not done
            if current_task is not None and not current_task.done():
                current_task.cancel()

            async def wait_and_invoke():
                await asyncio.sleep(wait)  # Delay execution
                await fn(*args, **kwargs)  # Invoke the async function

            # Schedule the new task
            current_task = asyncio.create_task(wait_and_invoke())

        return debounced
    return decorator


from pydantic import Field, validator
from zt_backend.models.components.zt_component import ZTComponent
from zt_backend.models.state.user_state import UserContext

class Timer(ZTComponent):
    """Timer is a component that allows for execution of code at a set interval. This does not have any visual output"""
    component: str = Field("v-timer", description="Vue component name")
    interval: int = Field(100000, description="Interval in milliseconds")
    value: bool = Field(False, description="Flag for execution under interval")
    triggerEvent: str = Field('click',description="Trigger event for when to execute a run")
        
    @validator('value', always=True) #TODO: debug and replace with field validator
    def get_value_from_global_state(cls, value, values):
        id = values['id'] # Get the id if it exists in the field values
        execution_state = UserContext.get_state()
        try:
            if execution_state and id and id in execution_state.component_values:  # Check if id exists in global_state
                return execution_state.component_values[id]  # Return the value associated with id in global_state
        except Exception as e:
            e
        return value  # If id doesn't exist in global_state, return the original value


from zt_backend.config import settings
from zt_backend.models.state.user_state import UserState
import asyncio
import logging
import traceback
import threading

logger = logging.getLogger("__name__")

class AppState:
    def __init__(self):
        self.user_states={}
        self.user_timers={}
        self.user_threads={}
        self.user_message_tasks={}
        self.current_thread=None
        self.notebook_state=UserState('')
        self.run_mode = settings.run_mode
        self.save_queue = asyncio.Queue()

    def stop_execution(self, data):
        if self.run_mode=='dev' and self.current_thread:
            self.current_thread.kill()
            self.notebook_state.current_cell_components.clear()
            self.notebook_state.current_cell_layout.clear()
            self.notebook_state.component_values.clear()
            self.notebook_state.created_components.clear()
            self.notebook_state.context_globals['exec_mode'] = False
        if self.run_mode=='app' and self.user_threads[data]:
            self.user_threads[data].kill()
            self.user_states[data].current_cell_components.clear()
            self.user_states[data].current_cell_layout.clear()
            self.user_states[data].component_values.clear()
            self.user_states[data].created_components.clear()
            self.user_states[data].context_globals['exec_mode'] = False

    def remove_user_state(self, user_id):
        try:
            if user_id in self.user_timers:
                # Cancel and remove the associated timer
                timer = self.user_timers[user_id]
                message_sender = self.user_message_tasks[user_id]
                if timer:
                    timer.cancel()
                del self.user_timers[user_id]
                if message_sender:
                    message_sender.cancel() 
                del self.user_message_tasks[user_id]
                if user_id in self.user_states: del self.user_states[user_id]
                logger.debug("User state removed for user %s", user_id)
        except Exception as e:
            logger.error("Error removing user state for user %s: %s", user_id, traceback.format_exc())

    def timer_set(self, user_id, timeout_seconds):
        logger.debug("Starting timer for user %s", user_id)
        if user_id in self.user_timers:
            existing_timer = self.user_timers[user_id]
            if existing_timer:
                existing_timer.cancel()
            
            timer = threading.Timer(timeout_seconds, self.remove_user_state, args=(user_id,))
            timer.daemon=True
            timer.start()
            
            self.user_timers[user_id] = timer

    def shutdown(self):
        if self.current_thread:
            self.current_thread.kill()
        for user_id in self.user_threads:
            if self.user_threads[user_id]:
                self.user_threads[user_id].kill()
        for user_id in self.user_timers:
            if self.user_timers[user_id]:
                self.user_timers[user_id].cancel()
        for user_id in self.user_message_tasks:
            if self.user_message_tasks[user_id]:
                self.user_message_tasks[user_id].cancel()


import threading
import sys

class KThread(threading.Thread):
  """A subclass of threading.Thread, with a kill()
method."""
  def __init__(self, *args, **keywords):
    threading.Thread.__init__(self, *args, **keywords)
    self.killed = False

  def start(self):
    """Start the thread."""
    self.__run_backup = self.run
    self.run = self.__run      # Force the Thread to install our trace.
    threading.Thread.start(self)

  def __run(self):
    """Hacked run function, which installs the
trace."""
    sys.settrace(self.globaltrace)
    self.__run_backup()
    self.run = self.__run_backup

  def globaltrace(self, frame, why, arg):
    if why == 'call':
      return self.localtrace
    else:
      return None

  def localtrace(self, frame, why, arg):
    if self.killed:
      if why == 'line':
        raise SystemExit()
    return self.localtrace

  def kill(self):
    self.killed = True


from typing import Optional
from pydantic import Field, validator
from zt_backend.models.components.zt_component import ZTComponent
from zt_backend.models.state.user_state import UserContext

class TextArea(ZTComponent):
    """Text area input allows a user to input arbitrary text. This is meant for larger text inputs"""
    component: str = Field("v-textarea", description="Vue component name")
    hint: Optional[str] = Field('Press Enter to Submit', description="Hint text for the text input")
    value: str = Field ('',description="The input text value")
    placeholder: Optional[str] = Field(None, description="Placeholder text")
    label: Optional[str] = Field(None, description="Label for the text input")
    readonly: Optional[bool] = Field(None, description="If true, the input is read-only")
    disabled: Optional[bool] = Field(None, description="If true, the input is disabled")
    triggerEvent: str = Field(None,description="Trigger event to send code to the backend")
    
    @validator('value', always=True) #TODO: debug and replace with field validator
    def get_value_from_global_state(cls, value, values):
        id = values['id'] # Get the id if it exists in the field values
        execution_state = UserContext.get_state()
        try:
            if execution_state and id and id in execution_state.component_values:  # Check if id exists in global_state
                return execution_state.component_values[id]  # Return the value associated with id in global_state
        except Exception as e:
            e
        return value  # If id doesn't exist in global_state, return the original value


from pydantic import Field
from typing import List, Optional
from zt_backend.models.components.zt_component import ZTComponent

class Card(ZTComponent):
    """A card is a container for components that should be displayed together. 
    Any child components will be placed in their own row within the card and take up the full width"""
    
    component: str = Field("v-card", description="Vue component name")
    childComponents: List[str] = Field([], description="List of child component ids to be placed within the card")
    color: Optional[str] = Field(None, description="Background color of the card")
    elevation: Optional[int] = Field(None, ge=0, le=24, description="Elevation level of the card. Must be between 0 and 24")
    density: Optional[str] = Field(None, enum=['default','comfortable','compact'], description="Density of the component")
    width: Optional[int] = Field(None, description="Width of the card")
    location: Optional[str] = Field(None, enum=['center'])


from zt_backend.models.api import request
import subprocess

def dependency_update(dependencyRequest: request.DependencyRequest):
    with open('requirements.txt', 'r+', encoding='utf-8') as file:
        contents = file.read()
        if contents == dependencyRequest.dependencies:
            return "No change to dependencies"
        file.seek(0)
        file.write(dependencyRequest.dependencies)
        file.truncate()
    subprocess.run(['pip', 'install', '-r', 'requirements.txt'])
    subprocess.run(['lock', 'requirements.txt'])


from zt_backend.models.api import request
from zt_backend.models import notebook
from typing import OrderedDict
import site
import uuid
import os

class NotebookState:
    def __init__(self):
        notebook_db_dir =  site.USER_SITE+'/.zero_true/'
        notebook_db_path = notebook_db_dir+'notebook.db'
        os.makedirs(notebook_db_dir, exist_ok=True)

        self.notebook_db_path = notebook_db_path
        codeCell = notebook.CodeCell(
                        id=str(uuid.uuid4()),
                        code='',
                        components=[],
                        variable_name='',
                        output='',
                        cellType='code'
                    )
        self.zt_notebook = notebook.Notebook(userId='', cells=OrderedDict([(codeCell.id, codeCell)]))
        self.base_cells = []
        self.base_components = {}


import subprocess
from fastapi import APIRouter, WebSocket, WebSocketDisconnect
from zt_backend.models import notebook
from zt_backend.models.api import request
from zt_backend.runner.execute_code import execute_request
from zt_backend.config import settings
from zt_backend.utils.completions import get_code_completions
from zt_backend.utils.dependencies import dependency_update
from zt_backend.utils.notebook import get_notebook_request, get_request_base, save_worker, websocket_message_sender
from zt_backend.models.managers.connection_manager import ConnectionManager
from zt_backend.models.managers.k_thread import KThread
from zt_backend.models.state.user_state import UserState
from zt_backend.models.state.app_state import AppState
from fastapi.responses import HTMLResponse
import logging
import uuid
import os
import traceback
import sys
import asyncio
import pkg_resources

router = APIRouter()
manager = ConnectionManager()
current_path = os.path.dirname(os.path.abspath(__file__))
app_state = AppState()

logger = logging.getLogger("__name__")

@router.get("/app", response_class=HTMLResponse)
async def catch_all():
    if(app_state.run_mode=='dev'):
        return HTMLResponse(open(os.path.join(current_path, "dist_dev", "index.html")).read())

@router.get("/health")
def health():
    return('UP')

@router.get("/env_data")
def env_data():
    return {
        "ws_url": settings.ws_url,
        "python_version": f'{sys.version_info.major}.{sys.version_info.minor}',
        "zt_version": pkg_resources.get_distribution('zero-true').version
    }

@router.get("/base_path")
def base_path():
    return settings.user_name + '/' + settings.project_name

@router.websocket("/ws/run_code")
async def run_code(websocket: WebSocket):
    if(app_state.run_mode=='dev'):
        message_send = asyncio.create_task(websocket_message_sender(app_state.notebook_state))
        await manager.connect(websocket)
        try:
            while True:
                data = await websocket.receive_json()
                app_state.notebook_state.websocket = websocket
                app_state.current_thread = KThread(target = execute_request, args=(request.Request(**data), app_state.notebook_state))
                app_state.current_thread.start()
        except WebSocketDisconnect:
            manager.disconnect(websocket)
        finally:
            message_send.cancel() 

@router.websocket("/ws/component_run")
async def component_run(websocket: WebSocket):
    await manager.connect(websocket)
    try:
        while True:
            data = await websocket.receive_json()
            logger.debug("Component change code execution started")
            component_request = request.ComponentRequest(**data)
            code_request = get_request_base(component_request.originId, component_request.components)
            if(app_state.run_mode=='dev'):
                app_state.notebook_state.websocket = websocket
                current_thread = KThread(target = execute_request, args=(code_request, app_state.notebook_state))
                current_thread.start()
            else:
                if component_request.userId not in app_state.user_states:
                    logger.debug("New user execution with id: %s, sending refresh", component_request.userId)
                    await websocket.send_json({"refresh": True})
                logger.debug("Existing user execution with id: %s", component_request.userId)
                app_state.timer_set(component_request.userId, 1800)
                app_state.user_states[component_request.userId].websocket = websocket
                app_state.user_threads[component_request.userId] = KThread(target = execute_request, args=(code_request, app_state.user_states[component_request.userId]))
                app_state.user_threads[component_request.userId].start()
    except WebSocketDisconnect:
        manager.disconnect(websocket)

@router.post("/api/create_cell")
def create_cell(cellRequest: request.CreateRequest):
     if(app_state.run_mode=='dev'):
        logger.debug("Code cell addition request started")
        createdCell = notebook.CodeCell(
            id=str(uuid.uuid4()),
            code='',
            components=[],
            output='',
            variable_name='',
            cellType=cellRequest.cellType
        )
        app_state.save_queue.put_nowait({"newCell": createdCell.model_copy(deep=True), "position_key":cellRequest.position_key})
        logger.debug("Code cell addition request completed")
        return createdCell

@router.post("/api/delete_cell")
def delete_cell(deleteRequest: request.DeleteRequest):
     cell_id = deleteRequest.cellId
     if(app_state.run_mode=='dev'):
        app_state.notebook_state.cell_outputs_dict.pop(cell_id, None)
        try:
            cell_dict = app_state.notebook_state.cell_outputs_dict['previous_dependecy_graph'].cells
            if cell_id in cell_dict:
                cell_dict.pop(cell_id, None)

            for cell_key, cell in cell_dict.items():
                if cell_id in dict(cell).get("child_cells", []):
                    cell["child_cells"].pop(cell_id, None)
                if cell_id in dict(cell).get("parent_cells", []):
                    cell["parent_cells"].pop(cell_id, None)

            logger.debug("Cell %s deleted successfully", cell_id)
        except Exception as e:
            logger.debug("Error when deleting cell %s: %s", cell_id, traceback.format_exc())
        app_state.save_queue.put_nowait({"deletedCell":cell_id})

@router.post("/api/hide_cell")
def hide_cell(hideCellRequest: request.HideCellRequest):
     if(app_state.run_mode=='dev'):
        logger.debug("Hide cell request started")
        app_state.save_queue.put_nowait({"hideCell": hideCellRequest})
        logger.debug("Hide cell request completed")

@router.post("/api/hide_code")
def hide_cell(hideCodeRequest: request.HideCodeRequest):
     if(app_state.run_mode=='dev'):
        logger.debug("Hide code request started")
        app_state.save_queue.put_nowait({"hideCode": hideCodeRequest})
        logger.debug("Hide code request completed")

@router.post("/api/rename_cell")
def rename_cell(renameCellRequest: request.NameCellRequest):
     if(app_state.run_mode=='dev'):
        logger.debug("Rename cell request started")
        app_state.save_queue.put_nowait({"renameCell": renameCellRequest})
        logger.debug("Rename cell request completed")

@router.post("/api/cell_reactivity")
def cell_reactivity(cellReactivityRequest: request.CellReactivityRequest):
     if(app_state.run_mode=='dev'):
        logger.debug("Cell reactivity request started")
        app_state.save_queue.put_nowait({"cellReactivity": cellReactivityRequest})
        logger.debug("Cell reactivity request completed")

@router.post("/api/expand_code")
def expand_code(expandCodeRequest: request.ExpandCodeRequest):
     if(app_state.run_mode=='dev'):
        logger.debug("Expand code request started")
        app_state.save_queue.put_nowait({"expandCode": expandCodeRequest})
        logger.debug("Expand code request completed")

@router.post("/api/show_table")
def show_table(showTableRequest: request.ShowTableRequest):
     if(app_state.run_mode=='dev'):
        logger.debug("Show Table request started")
        app_state.save_queue.put_nowait({"showTable": showTableRequest})
        logger.debug("Show Table request completed")

@router.websocket("/ws/save_text")
async def save_text(websocket: WebSocket):
    if(app_state.run_mode=='dev'):
        save_task = asyncio.create_task(save_worker(app_state.save_queue))
        await manager.connect(websocket)
        try:
            while True:
                data = await websocket.receive_json()
                cell_type = data.get("cellType")
                cell_id = data.get("id")
                app_state.save_queue.put_nowait({"saveCell": request.SaveRequest(id=cell_id, text=data.get("text"), cellType=cell_type)})
                if cell_type=="code":
                    completions = await get_code_completions(cell_id, data.get("code_w_context"), data.get("line"), data.get("column"))
                    await websocket.send_json(completions)
        except WebSocketDisconnect:
            manager.disconnect(websocket)
        finally:
            save_task.cancel()

@router.post("/api/clear_state")
def clear_state(clearRequest: request.ClearRequest):
     if(app_state.run_mode=='app'):
        logger.debug("Clearing state for user %s", clearRequest.userId)
        app_state.user_states.pop(clearRequest.userId, None)

@router.post("/api/dependency_update")
def dependency_update_request(dependencyRequest: request.DependencyRequest):
     if(app_state.run_mode=='dev'):
        logger.debug("Updating dependencies")
        try:
            dependency_update(dependencyRequest)
            logger.debug("Successfully updated dependencies")
        except Exception as e:
            logger.error('Error while updating requirements: %s', traceback.format_exc())

@router.websocket("/ws/notebook")
async def load_notebook(websocket: WebSocket):
    await manager.connect(websocket)
    try:
        while True:
            await websocket.receive_text()
            logger.debug("Get notebook request received")
            notebook_start = get_notebook_request()
            await websocket.send_json({"notebook_name": notebook_start.notebookName})
            if (app_state.run_mode=='app'):
                userId = str(uuid.uuid4())
                notebook_start.userId = userId
                app_state.user_states[userId]=UserState(userId)
                app_state.user_message_tasks[userId]=asyncio.create_task(websocket_message_sender(app_state.user_states[userId]))
                app_state.timer_set(userId, 1800)
                notebook_response = notebook.NotebookResponse(notebook=notebook_start, dependencies=notebook.Dependencies(value=''))
                await websocket.send_json(notebook_response.model_dump_json())
                app_state.user_states[userId].websocket = websocket
                app_state.user_threads[userId] = KThread(target = execute_request, args=(get_request_base(''), app_state.user_states[userId]))
                app_state.user_threads[userId].start()
            else:
                try:
                    with open('requirements.txt', 'r', encoding='utf-8') as file:
                        contents = file.read()
                    notebook_response =  notebook.NotebookResponse(notebook=notebook_start, dependencies=notebook.Dependencies(value=contents))
                    await websocket.send_json(notebook_response.model_dump_json())
                    await websocket.send_json({"complete": True})
                except FileNotFoundError:
                    logger.error('Requirements file not found')

    except WebSocketDisconnect:
        manager.disconnect(websocket)

@router.websocket("/ws/stop_execution")
async def stop_execution(websocket: WebSocket):
    await manager.connect(websocket)
    try:
        while True:
            data = await websocket.receive_text()
            app_state.stop_execution(data)
    except WebSocketDisconnect:
        manager.disconnect(websocket)

@router.post("/api/notebook_name_update")
def notebook_name_update(notebook_name: request.NotebookNameRequest):
    if(app_state.run_mode=='dev'):
        app_state.save_queue.put_nowait({"new_notebook_name":notebook_name.notebookName})

@router.post("/api/share_notebook")
def share_notebook(shareRequest: request.ShareRequest):
    if(app_state.run_mode=='dev'):
        subprocess.run(['zero-true', 'publish', shareRequest.apiKey, shareRequest.userName, shareRequest.projectName, '.'])

@router.on_event('shutdown')
def shutdown():
    app_state.shutdown()



from pydantic import Field, validator, field_validator
from zt_backend.models.components.zt_component import ZTComponent
from zt_backend.models.components.validations import validate_color
from typing import List, Optional, Union
from zt_backend.models.state.user_state import UserContext

class Autocomplete(ZTComponent):
    """Autocomplete is a select box that allows users to search the available options. 
    Optimal for when there are many options to choose from"""
    
    component: str = Field("v-autocomplete", description="Vue component name")
    items: List[Union[str, int]] = Field(..., description="Options for the autocomplete box. Can be a list of strings or integers")
    value: Union[str, int, None] = Field("", description="Selected option for the autocomplete box")
    label: Optional[str] = Field(None, description="Label for the autocomplete box")
    multiple: Optional[bool] = Field(None, description="Determines if multiple selections are allowed")
    clearable: Optional[bool] = Field(None, description="Determines if the autocomplete box has a clearable option")
    disabled: Optional[bool] = Field(None, description="Determines if the autocomplete box is disabled")
    readonly: Optional[bool] = Field(None, description="Determines if the autocomplete box is read-only")
    color: Optional[str] = Field(None, pre=True, description="Color of the autocomplete component. Can be custom or standard Material color")
    triggerEvent: str = Field('update:modelValue',description="Trigger event for when to run based on the selected value")

    @field_validator('color')
    def validate_color(cls, color):
        return validate_color(color)

    @validator('value', always=True) #TODO: debug and replace with field validator
    def get_value_from_global_state(cls, value, values):
        id = values['id'] # Get the id if it exists in the field values
        execution_state = UserContext.get_state()
        try:
            if execution_state and id and id in execution_state.component_values:  # Check if id exists in global_state
                return execution_state.component_values[id]  # Return the value associated with id in global_state
        except Exception as e:
            e
        return value  # If id doesn't exist in global_state, return the original value


import pandas as pd
from pydantic import Field, BaseModel
import numpy as np
from zt_backend.models.components.zt_component import ZTComponent
from typing import List, Dict,Any


class Header(BaseModel):
    """Header class for the columns of a DataFrame component"""
    title: str = Field("", description="Title of the column")
    align: str = Field("start", description="Alignment of values in the column")
    key: str = Field("name", description="Key of the column, must match the key in the items list")

class DataFrame(ZTComponent):
    """DataFrame component for displaying tabluar data"""
    component: str = Field("v-data-table", description="Vue component name.")
    headers: List[Header] = Field([], description="List of column headers for the DataFrame")
    items: List[Dict[str, Any]] = Field([], description="List of items to be displayed in the DataFrame")

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, id: str):
        """Create a DataFrame component from a pandas DataFrame"""
        df = df.replace({np.nan:None}).replace({np.inf:None}).replace({-np.inf:None})
        headers = [{"title": col, "key": col} for col in df.columns]
        items = df.to_dict(orient='records')
        return cls(id=id, headers=headers, items=items)